# Generated by Django 3.0.5 on 2020-08-24 16:24

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('quartet_output', '0004_auto_20190725_1122'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticationinfo',
            name='description',
            field=models.CharField(blank=True, help_text='An optional description.', max_length=200, null=True, verbose_name='Description'),
        ),
        migrations.AlterField(
            model_name='authenticationinfo',
            name='private_key',
            field=models.TextField(blank=True, help_text='Any private key value if applicable.', null=True, verbose_name='Private Key'),
        ),
        migrations.AlterField(
            model_name='authenticationinfo',
            name='public_key',
            field=models.TextField(blank=True, help_text='Any public key info if applicable.', null=True, verbose_name='Public Key'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='action',
            field=models.CharField(blank=True, choices=[('ADD', 'Add'), ('OBSERVE', 'Observe'), ('DELETE', 'Delete')], help_text="The EPCIS event's ACTION type.", max_length=20, null=True, verbose_name='Action'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='authentication_info',
            field=models.ForeignKey(blank=True, help_text='The Authentication Info to use.', null=True, on_delete=django.db.models.deletion.SET_NULL, to='quartet_output.AuthenticationInfo', verbose_name='Authentication Info'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='biz_location',
            field=models.CharField(blank=True, help_text='The business location URN.  Typically representing a site', max_length=150, null=True, verbose_name='Business Location'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='biz_step',
            field=models.CharField(blank=True, help_text='The business step URN.  Can be a CBV value or any customuri.  If CBV it must be exactly as specified in v1.2', max_length=150, null=True, verbose_name='Business Step (BizStep)'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='destination_id',
            field=models.CharField(blank=True, help_text='A URI that identifies the Destination specified in the Destination Type field.', max_length=200, null=True, verbose_name='Destination ID'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='destination_type',
            field=models.CharField(blank=True, help_text='The type of the Destination- a CBV 1.2 URI or custom URI.', max_length=150, null=True, verbose_name='Destination Type'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='disposition',
            field=models.CharField(blank=True, help_text='A Disposition URN- can be CBV or custom.', max_length=150, null=True, verbose_name='Disposition'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='event_type',
            field=models.CharField(blank=True, choices=[('Object', 'Object'), ('Transaction', 'Transaction'), ('Transformation', 'Transformation'), ('Aggregation', 'Aggregation')], help_text='The type of EPCIS event.', max_length=20, null=True, verbose_name='Event Type'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='read_point',
            field=models.CharField(blank=True, help_text='The read point URN.  Typically representing a sub-site.', max_length=150, null=True, verbose_name='Read Point'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='receiver_identifier',
            field=models.CharField(blank=True, help_text='Typically an SGLN but an identifier that is in the SBDH and uniquely identifies a receiving entity.', max_length=250, null=True, verbose_name='SBDH Receiver Identifier'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='sender_identifier',
            field=models.CharField(blank=True, help_text='Typically an SGLN but an identifier that is in the SBDH and uniquely identifies a sending entity.', max_length=250, null=True, verbose_name='SBDH Sender Identifier'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='source_id',
            field=models.CharField(blank=True, help_text='A URI that identifies the source specified in the Source Type field.', max_length=200, null=True, verbose_name='Source ID'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='source_type',
            field=models.CharField(blank=True, help_text='The type of the source- a CBV 1.2 URI or custom URI.', max_length=150, null=True, verbose_name='Source Type'),
        ),
    ]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2019 SerialLab Corp.  All rights reserved.
import socket
from quartet_output.models import EPCISOutputCriteria
from quartet_capture.rules import RuleContext
from urllib.parse import urlparse

# mysocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
# host = "10.80.209.106"
# port = 9100
# try:
#     mysocket.connect((host, port))  # connecting to host
#     mysocket.send(b"^XA^A0N,50,50^FO50,50^FDSocket Test^FS^XZ")  # using bytes
#     mysocket.close()  # closing connection
# except:
#     print("Error with the connection")


class SocketTransportMixin:

    def socket_send(self, data, rule_context: RuleContext,
                    output_criteria: EPCISOutputCriteria,
                    info):
        """
        Sends data over a socket
        :param data: The data to send (as bytes)
        :param rule_context: The rule context
        :param output_criteria: The output criteria supplying the urn
        :param info: The info function for logging
        :return: None
        """
        mysocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        parsed_url = urlparse(output_criteria.end_point.urn)
        host = parsed_url.netloc.split(":")[0]
        port = parsed_url.port

        info('Connecting to host %s ', output_criteria.end_point.urn)
        mysocket.connect((host, port))  # connecting to host
        info('Sending data: %s (may be truncated for display purposes)',
                  data[:1000])
        mysocket.send(
            bytes(data))  # using bytes
        mysocket.close()  # closing connection



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
from django.utils.translation import gettext as _
from django.core.management.base import BaseCommand, CommandError
from quartet_output.management.commands.utils import create_output_filter_rule


class Command(BaseCommand):
    help = _(
        'Loads the quartet_output demonstration and transport rules into '
        'the database.'
    )

    def handle(self, *args, **options):
        create_output_filter_rule()
        create_output_filter_rule(rule_name='EPCIS Delayed Output Filter',
                                  delay_rule=True)




# Generated by Django 2.0.5 on 2019-07-25 16:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('quartet_output', '0003_auto_20190214_1806'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='authenticationinfo',
            options={'ordering': ['username'], 'verbose_name': 'Authentication Info', 'verbose_name_plural': 'Authentication Info'},
        ),
        migrations.AlterModelOptions(
            name='endpoint',
            options={'ordering': ['name'], 'verbose_name': 'End Point', 'verbose_name_plural': 'End Points'},
        ),
        migrations.AlterModelOptions(
            name='epcisoutputcriteria',
            options={'ordering': ['name'], 'verbose_name': 'EPCIS Output Criteria', 'verbose_name_plural': 'EPCIS Output Criteria'},
        ),
    ]



# Generated by Django 2.0.5 on 2018-09-09 14:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('quartet_output', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='epcisoutputcriteria',
            name='receiver_identifier',
            field=models.CharField(help_text='Typically an SGLN but an identifier that is in the SBDH and uniquely identifies a receiving entity.', max_length=250, null=True, verbose_name='SBDH Receiver Identifier'),
        ),
        migrations.AddField(
            model_name='epcisoutputcriteria',
            name='sender_identifier',
            field=models.CharField(help_text='Typically an SGLN but an identifier that is in the SBDH and uniquely identifies a sending entity.', max_length=250, null=True, verbose_name='SBDH Sender Identifier'),
        ),
        migrations.AlterField(
            model_name='endpoint',
            name='name',
            field=models.CharField(help_text='The name of the endpoint', max_length=150, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='epcisoutputcriteria',
            name='name',
            field=models.CharField(help_text='The name of the criteria record', max_length=150, unique=True, verbose_name='Name'),
        ),
    ]



import os

from django.test import TestCase

from EPCPyYes.core.v1_2.CBV.business_steps import BusinessSteps
from EPCPyYes.core.v1_2.CBV.dispositions import Disposition
from EPCPyYes.core.v1_2.events import EventType
from quartet_capture.models import Rule, Step, StepParameter, Task
from quartet_capture.tasks import execute_rule
from quartet_epcis.parsing.context_parser import BusinessEPCISParser
from quartet_output import models
from quartet_output.models import EPCISOutputCriteria
from quartet_output.steps import ContextKeys


class TestOutputParsing(TestCase):
    def _create_endpoint(self):
        ep = models.EndPoint()
        ep.urn = 'mailto:crashtestdummie@unittest.local?body=send%20current-issue&subject=awesome email'
        ep.name = 'Test EndPoint'
        ep.save()
        return ep

    def _create_good_ouput_criterion(self):
        endpoint = self._create_endpoint()
        eoc = EPCISOutputCriteria()
        eoc.name = "Test Criteria"
        eoc.action = "ADD"
        eoc.event_type = EventType.Aggregation.value
        eoc.disposition = Disposition.in_progress.value
        eoc.biz_step = BusinessSteps.packing.value
        eoc.read_point = 'urn:epc:id:sgln:0555555.00002.0'
        eoc.end_point = endpoint
        eoc.save()
        return eoc

    def _create_step(self, rule):
        step = Step()
        step.rule = rule
        step.order = 1
        step.name = 'Output Determination'
        step.step_class = 'quartet_output.steps.OutputParsingStep'
        step.description = 'unit test step'
        step.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'EPCIS Output Criteria'
        step_parameter.value = 'Test Criteria'
        step_parameter.save()
        return step

    def _create_comm_step(self, rule):
        step = Step()
        step.rule = rule
        step.order = 3
        step.name = 'CreateCommissioning'
        step.step_class = 'quartet_output.steps.AppendCommissioningStep'
        step.description = 'unit test commissioning step'
        step.save()

    def _create_epcpyyes_step(self, rule, json=False):
        step = Step()
        step.rule = rule
        step.order = 4
        step.name = 'Create EPCIS'
        step.step_class = 'quartet_output.steps.EPCPyYesOutputStep'
        step.description = 'Creates EPCIS XML or JSON and inserts into rule' \
                           'context.'
        step.save()
        if json:
            param = StepParameter.objects.create(
                step=step,
                name='JSON',
                value=True
            )

    def _create_task_step(self, rule, order=5):
        step = Step()
        step.rule = rule
        step.order = order
        step.name = 'Create Output Task'
        step.step_class = 'quartet_output.steps.CreateOutputTaskStep'
        step.description = 'Looks for any EPCIS data on the context and ' \
                           'then, if found, creates a new output task using ' \
                           'the configured Output Rule step parameter.'
        step.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'Output Rule'
        step_parameter.value = 'Transport Rule'
        step_parameter.description = 'The name of the rule to create a new ' \
                                     'task with.'
        step_parameter.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'Run Immediately'
        step_parameter.value = 'True'
        step_parameter.description = 'The name of the rule to create a new ' \
                                     'task with.'
        step_parameter.save()
        return step

    def _create_rule(self):
        rule = Rule()
        rule.name = 'output-test'
        rule.description = 'output test rule'
        rule.save()
        return rule

    def _create_transport_rule(self):
        rule = Rule()
        rule.name = 'Transport Rule'
        rule.description = 'Attempts to send data using transport step(s).'
        rule.save()
        return rule

    def _create_transport_step(self, rule, put_data='False'):
        step = Step()
        step.rule = rule
        step.order = 1
        step.name = 'Transport'
        step.step_class = 'quartet_output.steps.TransportStep'
        step.description = 'Sends test data.'
        step.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'run-immediately'
        step_parameter.value = 'True'
        step_parameter.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'put-data'
        step_parameter.value = put_data
        step_parameter.save()

    def _create_task(self, rule):
        task = Task()
        task.rule = rule
        task.name = 'unit test task'
        task.save()
        return task

    def _parse_test_data(self, test_file='data/epcis.xml',
                         parser_type=BusinessEPCISParser,
                         recursive_decommission=False):
        curpath = os.path.dirname(__file__)
        if isinstance(parser_type, BusinessEPCISParser):
            parser = BusinessEPCISParser(
                os.path.join(curpath, test_file),
                recursive_decommission=recursive_decommission
            )
        else:
            parser = parser_type(
                os.path.join(curpath, test_file),
            )
        message_id = parser.parse()
        print(parser.event_cache)
        parser.clear_cache()
        return message_id, parser

    def test_rule_with_agg_comm_output(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        self._create_task_step(db_rule)
        db_rule2 = self._create_transport_rule()
        self._create_transport_step(db_rule2)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commissioning.xml')
        data_path = os.path.join(curpath, 'data/aggregation.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.FILTERED_EVENTS_KEY.value]),
                12,
                "There should be twelve filtered events."
            )
            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY.value]
            task = Task.objects.get(name=task_name)
            self.assertEqual(task.status, 'FINISHED')
            print(context.context[ContextKeys.OUTBOUND_EPCIS_MESSAGE_KEY.value])





# Generated by Django 2.0.5 on 2018-08-08 13:16

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AuthenticationInfo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('username', models.CharField(help_text='The username.', max_length=150, verbose_name='Username')),
                ('password', models.CharField(help_text='The password for the user.', max_length=25, verbose_name='Password')),
                ('type', models.CharField(help_text='The type of authentication to use.  For example, Basic, Digest, SSH, etc.  This is optional depending on whether or not any steps require this information.', max_length=25, verbose_name='Type')),
                ('description', models.CharField(help_text='An optional description.', max_length=200, null=True, verbose_name='Description')),
                ('private_key', models.TextField(help_text='Any private key value if applicable.', null=True, verbose_name='Private Key')),
                ('public_key', models.TextField(help_text='Any public key info if applicable.', null=True, verbose_name='Public Key')),
            ],
        ),
        migrations.CreateModel(
            name='EndPoint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of the endpoint', max_length=150, verbose_name='Name')),
                ('urn', models.CharField(help_text='A uniform resource locator that includes protocol information.', max_length=1000, verbose_name='URL')),
            ],
        ),
        migrations.CreateModel(
            name='EPCISOutputCriteria',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of the endpoint', max_length=150, verbose_name='Name')),
                ('event_type', models.CharField(choices=[('Object', 'Object'), ('Transaction', 'Transaction'), ('Transformation', 'Transformation'), ('Aggregation', 'Aggregation')], help_text='The type of EPCIS event.', max_length=20, null=True, verbose_name='Event Type')),
                ('action', models.CharField(choices=[('ADD', 'Add'), ('OBSERVE', 'Observe'), ('DELETE', 'Delete')], help_text="The EPCIS event's ACTION type.", max_length=20, null=True, verbose_name='Action')),
                ('biz_step', models.CharField(help_text='The business step URN.  Can be a CBV value or any customuri.  If CBV it must be exactly as specified in v1.2', max_length=150, null=True, verbose_name='Business Step (BizStep)')),
                ('disposition', models.CharField(help_text='A Disposition URN- can be CBV or custom.', max_length=150, null=True, verbose_name='Disposition')),
                ('read_point', models.CharField(help_text='The read point URN.  Typically representing a sub-site.', max_length=150, null=True, verbose_name='Read Point')),
                ('biz_location', models.CharField(help_text='The business location URN.  Typically representing a site', max_length=150, null=True, verbose_name='Business Location')),
                ('source_type', models.CharField(help_text='The type of the source- a CBV 1.2 URI or custom URI.', max_length=150, null=True, verbose_name='Source Type')),
                ('source_id', models.CharField(help_text='A URI that identifies the source specified in the Source Type field.', max_length=200, null=True, verbose_name='Source ID')),
                ('destination_type', models.CharField(help_text='The type of the Destination- a CBV 1.2 URI or custom URI.', max_length=150, null=True, verbose_name='Destination Type')),
                ('destination_id', models.CharField(help_text='A URI that identifies the Destination specified in the Destination Type field.', max_length=200, null=True, verbose_name='Destination ID')),
                ('authentication_info', models.ForeignKey(help_text='The Authentication Info to use.', null=True, on_delete=django.db.models.deletion.SET_NULL, to='quartet_output.AuthenticationInfo', verbose_name='Authentication Info')),
                ('end_point', models.ForeignKey(help_text='A prtocol-specific endpoint defining where any output data will be sent.', on_delete=django.db.models.deletion.PROTECT, to='quartet_output.EndPoint', verbose_name='End Point')),
            ],
            options={
                'verbose_name': 'EPCIS Output Criteria',
                'verbose_name_plural': 'EPCIS Output Criteria',
            },
        ),
    ]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
from django.utils.translation import gettext as _
from django.db.utils import IntegrityError
from quartet_capture import models
from quartet_output.models import EndPoint, EPCISOutputCriteria, \
    AuthenticationInfo


def create_output_filter_rule(rule_name='EPCIS Output Filter',
                              delay_rule=False):
    create_criteria()
    if not models.Rule.objects.filter(name=rule_name).exists():
        rule = models.Rule.objects.create(
            name=rule_name,
            description=_('Will inspect inbound messages for output '
                          'processing.')
        )
        parse_step = models.Step.objects.create(
            name=_('Inspect EPCIS'),
            description=_(
                'Parse and insepect EPCIS events using output criteria.'),
            step_class='quartet_output.steps.OutputParsingStep',
            order=1,
            rule=rule
        )
        models.StepParameter.objects.create(
            name='EPCIS Output Criteria',
            step=parse_step,
            value='Test Transaction Criteria',
            description=_(
                'This is the name of the EPCIS Output Criteria record to use.')

        )
        models.Step.objects.create(
            name=_('Add Commissioning Data'),
            description=_(
                'Adds commissioning events for filtered EPCs and their children.'),
            step_class='quartet_output.steps.AddCommissioningDataStep',
            order=2,
            rule=rule
        )
        models.Step.objects.create(
            name=_('Add Aggregation Data'),
            description=_(
                'Adds aggregation events for included EPCs in any filtered events.'),
            step_class='quartet_output.steps.UnpackHierarchyStep',
            order=3,
            rule=rule
        )
        models.Step.objects.create(
            name=_('Render EPCIS XML'),
            description=_(
                'Pulls any EPCPyYes objects from the context and creates an XML message'),
            step_class='quartet_output.steps.EPCPyYesOutputStep',
            order=4,
            rule=rule
        )
        output_step = models.Step.objects.create(
            name=_('Queue Outbound Message'),
            description=_('Creates a Task for sending any outbound data'),
            step_class='quartet_output.steps.CreateOutputTaskStep',
            order=5,
            rule=rule
        )
        models.StepParameter.objects.create(
            step=output_step,
            name='Output Rule',
            value='Transport Rule'
        )
        if delay_rule == True:
            second_parse_step = models.Step.objects.create(
                name=_('Inspect EPCIS'),
                description='Used to assign an output criteria to the second '
                            'message for forwarding filtered events.',
                step_class='quartet_output.steps.OutputParsingStep',
                order=6,
                rule=rule
            )
            models.StepParameter.objects.create(
                name='EPCIS Output Criteria',
                step=second_parse_step,
                value='Test Transaction Criteria',
                description=_(
                    'This is the name of the EPCIS Output Criteria record to use.')

            )
            render_events = models.Step.objects.create(
                name=_('Render Filtered Events'),
                description=_('Takes any events that were filtered by the '
                              'prior step and renders them to XML for '
                              'sending by the next step.'),
                step_class='quartet_output.steps.EPCPyYesFilteredEventOutputStep',
                order=7,
                rule=rule
            )
            queue_outbound_message = models.Step.objects.create(
                name='Queue Outbound Message',
                description=_('Creates a task and sends it to the delayed '
                              'transport rule or whatever transport rule is '
                              'configured via the Output Rule step parameter.'),
                step_class='quartet_output.steps.CreateOutputTaskStep',
                order=8,
                rule=rule
            )
            output_rule_param = models.StepParameter.objects.create(
                name='Output Rule',
                description=_('The name of the rule to send the context '
                              'data to.'),
                value='Delayed Transport Rule',
                step=queue_outbound_message
            )
            sdstep = create_transport_rule(rule_name='Delayed Transport Rule',
                                           add_delay=delay_rule)
        sdstep = create_transport_rule()
        return rule

def create_transport_rule(rule_name='Transport Rule', add_delay=False):
    try:
        trule = models.Rule.objects.create(
            name=rule_name,
            description=_(
                'An output Rule for any data filtered by EPCIS Output Criteria '
                'rules.')
        )
        if add_delay:
            delay_step = models.Step.objects.create(
                name=_('Wait Three Seconds'),
                description=_(
                    "Wait for three seconds until moving to the next "
                    "step"),
                step_class='quartet_output.steps.DelayStep',
                rule=trule,
                order=1
            )
            delay_step_param = models.StepParameter.objects.create(
                name='Timeout Interval',
                value='3',
                description=_(
                    'The amount of time in seconds to pause the rule.'),
                step=delay_step
            )

        sdstep = models.Step.objects.create(
            name=_('Send Data'),
            description=_(
                'This will send the task message using the source EPCIS Output '
                'Critria EndPoint and Authentication Info.'),
            step_class='quartet_output.steps.TransportStep',
            order=2,
            rule=trule
        )
    except IntegrityError:
        trule = models.Rule.objects.get(name=rule_name)
    return trule


def create_criteria(endpoint_name='Local Echo Server',
                    username=None,
                    criteria_name=None):
    try:
        endpoint = EndPoint.objects.create(
            name=_(endpoint_name),
            urn=_('http://localhost')
        )
    except IntegrityError:
        print('Endpoint already exists.')
        endpoint = EndPoint.objects.get(name=endpoint_name)
    try:
        auth = AuthenticationInfo.objects.create(
            username=_('Test User') or username,
            password=_('Test Password'),
            type='Digest',
            description=_('A test user for the example rule.'))
    except IntegrityError:
        print('Authentication info already exists.')
        auth = AuthenticationInfo.objects.get(username='Test User')
    try:
        output = EPCISOutputCriteria.objects.create(
            name=_('Test Transaction Criteria') or criteria_name,
            action='ADD',
            event_type='Transaction',
            biz_location='urn:epc:id:sgln:305555.123456.0',
            end_point=endpoint,
            authentication_info=auth
        )
    except IntegrityError:
        print('Criteria already exists.')



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2019 SerialLab Corp.  All rights reserved.
from urllib.parse import urlparse, parse_qsl

from django.core.mail import EmailMessage

from quartet_capture.rules import RuleContext
from quartet_output.models import EPCISOutputCriteria


class MailMixin:
    '''
    Uses default django functionality to send an email.
    Emails are sent by configuring mailto links in the
    quartet endpoint configuration and attaching to output criteria.

    Mailto endpoints require no corresponding authentication info classes
    since only one global SMTP server per QU4RTET instance can be defined
    in the settings for sending outbound messages.
    '''
    def send_email(self, data: str, rule_context: RuleContext,
                   output_criteria: EPCISOutputCriteria,
                   info_func,
                   file_extension='txt',
                   mimetype='text/plain'):
        '''
        Parses out a mailto link and sends the data parameter to the addresses
        specified in the link.

        :param data: The data to send.
        :param rule_context: The quartet capture rules.RuleContext instance
            from the currently running rule.
        :param output_criteria: The models.OutputCriteria instance from the
            current TransportStep being executed.
        :param info_func: The info logging function from the calling step
            class.
        :param file_extension: This is the file extension for the attachement
            being sent.  It is best to leave it as txt even if the "real" data is
            JSON or XML since many email filters will block those formats.
        :param mimetype: The mimetype of the attachment.  Default is text/plain.
        :return: None.
        '''
        email = self.convert_mailto_url(output_criteria.end_point.urn)
        filename = '%s.%s' % (rule_context.task_name, file_extension)
        email.attach(filename, data, mimetype=mimetype)
        email.send()

    def convert_mailto_url(self, mailto):
        mail_info = urlparse(mailto, 'mailto')
        if mail_info.query:
            query_dict = dict(parse_qsl(mail_info.query))
        else:
            query_dict = {
                'subject': 'QU4RTET Output Messaging',
                'body': 'The included attachment has been sent from '
                        'a QU4RTET system to your inbox.',
                'reply_to': 'donotreply@qu4rtet.io',
                'from_email': 'output@qu4rtet.io'
            }
        cc = None
        bcc = None
        if 'cc' in query_dict.keys():
            cc = query_dict['cc'].split(',')
        if 'bcc' in query_dict.keys():
            bcc = query_dict['bcc'].split(',')
        email = EmailMessage(
            subject=query_dict.get('subject'),
            to=mail_info.path.split(','),
            body=query_dict.get('body', None),
            cc=cc,
            bcc=bcc,
            reply_to=query_dict.get('reply-to')
        )
        return email



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_quartet_output
------------

Tests for `quartet_output` models module.
"""
import os
from urllib.parse import urlparse

import paramiko
import time
from django.conf import settings
from django.core.exceptions import ValidationError
from django.test import TestCase

from EPCPyYes.core.SBDH.sbdh import StandardBusinessDocumentHeader as sbdheader
from EPCPyYes.core.v1_2.CBV.business_steps import BusinessSteps
from EPCPyYes.core.v1_2.CBV.dispositions import Disposition
from EPCPyYes.core.v1_2.events import EventType
from quartet_capture.models import Rule, Step, StepParameter, Task
from quartet_capture.tasks import execute_rule, execute_queued_task
from quartet_epcis.parsing.context_parser import BusinessEPCISParser
from quartet_output import models
from quartet_output.models import EPCISOutputCriteria
from quartet_output.steps import SimpleOutputParser, ContextKeys
from quartet_output.transport.mail import MailMixin


class TestQuartetOutput(TestCase):

    def test_bad_source(self):
        oc = self._create_good_ouput_criterion()
        oc.source_id = None
        with self.assertRaises(ValidationError):
            oc.clean()

    def test_bad_destination(self):
        oc = self._create_good_ouput_criterion()
        oc.destination_id = None
        with self.assertRaises(ValidationError):
            oc.clean()

    def test_rule_with_pallet_agg(self):
        '''
        Should process the EPCIS but not filter out a full pallet hierarchy.
        :return:
        '''
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_three_events.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)

    def test_rule_with_agg_comm(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_one_event.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            self.assertIsNotNone(
                context.context.get(
                    ContextKeys.EPCIS_OUTPUT_CRITERIA_KEY.value)
            )

    def test_rule_with_forward_data(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_forward_data_step(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_one_event.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            self.assertIsNotNone(
                context.context.get(
                    ContextKeys.EPCIS_OUTPUT_CRITERIA_KEY.value)
            )

    def test_rule_with_agg_trigger(self):
        self._create_good_agg_trigger_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        self._parse_test_data('data/commissioning.xml')
        self._parse_test_data('data/aggregation.xml')

    def test_rule_with_agg_comm_json_output(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule, json=True)
        self._create_task_step(db_rule)
        db_rule2 = self._create_transport_rule()
        self._create_transport_step(db_rule2)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_one_event.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY.value]
            execute_queued_task(task_name=task_name)
            task = Task.objects.get(name=task_name)
            self.assertEqual(task.status, 'FINISHED')

    def test_rule_with_agg_comm_output(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        self._create_task_step(db_rule)
        db_rule2 = self._create_transport_rule()
        self._create_transport_step(db_rule2)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_one_event.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY.value]
            execute_queued_task(task_name=task_name)
            task = Task.objects.get(name=task_name)
            self.assertEqual(task.status, 'FINISHED')

    def test_rule_with_agg_comm_output_delay(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        self._create_task_step(db_rule)
        self.create_delay_step(db_rule, order=10)
        self._create_filtered_output_step(db_rule, order=20)
        db_rule2 = self._create_transport_rule()
        self._create_transport_step(db_rule2)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_one_event.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY.value]
            execute_queued_task(task_name=task_name)
            task = Task.objects.get(name=task_name)
            self.assertEqual(task.status, 'FINISHED')

    def test_rule_with_header_output(self):
        self._create_good_header_criterion()
        db_rule = self._create_rule()
        db_step = self._create_step(db_rule)
        db_task_step = self._create_task_step(db_rule)
        self._add_forward_data_step_parameter(db_task_step)
        db_step.order = 1
        db_task_step.order = 2
        db_step.save()
        db_task_step.save()
        db_rule2 = self._create_transport_rule()
        self._create_transport_step(db_rule2)
        curpath = os.path.dirname(__file__)
        db_task = self._create_task(db_rule)
        data_path = os.path.join(curpath, 'data/epcis.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            for event in context.context[
                ContextKeys.FILTERED_EVENTS_KEY.value
            ]:
                self.assertIsInstance(event, sbdheader)
            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY.value]
            execute_queued_task(task_name=task_name)
            task = Task.objects.get(name=task_name)
            self.assertEqual(task.status, 'FINISHED')

    def test_rule_with_agg_comm_output_put(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        self._create_task_step(db_rule)
        db_rule2 = self._create_transport_rule()
        self._create_transport_step(db_rule2, put_data=True)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_one_event.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY.value]
            execute_queued_task(task_name=task_name)
            task = Task.objects.get(name=task_name)
            self.assertEqual(task.status, 'FINISHED')

    def test_rule_with_agg_mulit_comm(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_three_events.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            self.assertIn(ContextKeys.OUTBOUND_EPCIS_MESSAGE_KEY.value,
                          context.context,
                          "No EPCIS message was found in the rule context.")
            self.assertIn(ContextKeys.OBJECT_EVENTS_KEY.value,
                          context.context)
            self.assertEqual(
                len(context.context[ContextKeys.OBJECT_EVENTS_KEY.value]), 3,
                "There should be three object events."
            )
            for oevent in context.context[
                ContextKeys.OBJECT_EVENTS_KEY.value]:
                self.assertIn(len(oevent.epc_list), [10, 2, 1],
                              "One of the object events in the context is "
                              "malformed.")

    def test_rule_no_agg(self):
        '''
        Should process the EPCIS but not filter out any aggregation events.
        :return:
        '''
        # create output criteria that does not result in the
        # matching of any events.
        self._create_bad_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        data_path = os.path.join(curpath, 'data/epcis.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                0,
                "There should be no filtered events."
            )

    def test_full_bad_rule(self):
        oec = self._create_good_ouput_criterion()
        oec.name = "Bad Name"
        oec.save()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        data_path = os.path.join(curpath, 'data/epcis.xml')
        with open(data_path, 'r') as data_file:
            try:
                execute_rule(data_file.read().encode(), db_task)
            except EPCISOutputCriteria.DoesNotExist:
                pass

    def test_delay_rule(self):
        start = time.time()
        db_rule = self._create_delay_rule()
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        data_path = os.path.join(curpath, 'data/epcis.xml')
        with open(data_path, 'r') as data_file:
            try:
                execute_rule(data_file.read().encode(), db_task)
            except EPCISOutputCriteria.DoesNotExist:
                pass
        stop = time.time()
        self.assertGreaterEqual((stop - start), 3, 'The rule did not pause '
                                                   'long enough.')

    def _create_good_ouput_criterion(self):
        endpoint = self._create_endpoint()
        auth = self._create_auth()
        eoc = EPCISOutputCriteria()
        eoc.name = "Test Criteria"
        eoc.action = "ADD"
        eoc.event_type = EventType.Transaction.value
        eoc.disposition = Disposition.in_transit.value
        eoc.biz_step = BusinessSteps.shipping.value
        eoc.biz_location = 'urn:epc:id:sgln:305555.123456.0'
        eoc.read_point = 'urn:epc:id:sgln:305555.123456.12'
        eoc.source_type = 'urn:epcglobal:cbv:sdt:location'
        eoc.source_id = 'urn:epc:id:sgln:305555.123456.12'
        eoc.destination_type = 'urn:epcglobal:cbv:sdt:location'
        eoc.destination_id = 'urn:epc:id:sgln:309999.111111.233'
        eoc.authentication_info = auth
        eoc.end_point = endpoint
        eoc.save()
        return eoc

    def _create_good_agg_trigger_ouput_criterion(self):
        endpoint = self._create_endpoint()
        auth = self._create_auth()
        eoc = EPCISOutputCriteria()
        eoc.name = "Test Criteria"
        eoc.action = "ADD"
        eoc.biz_step = BusinessSteps.packing.value
        eoc.biz_location = 'urn:epc:id:sgln:0555555.00002.0'
        eoc.sender_identifier = 'urn:epc:id:sgln:0555555.00001.0'
        eoc.end_point = endpoint
        eoc.save()
        return eoc

    def _create_good_header_criterion(self):
        endpoint = self._create_endpoint()
        auth = self._create_auth()
        eoc = EPCISOutputCriteria()
        eoc.name = "Test Criteria"
        eoc.receiver_identifier = 'urn:epc:id:sgln:039999.111111.0'
        eoc.authentication_info = auth
        eoc.end_point = endpoint
        eoc.save()
        return eoc

    def _create_good_header_criterion(self):
        eoc = EPCISOutputCriteria()
        eoc.name = 'Test Criteria'
        eoc.receiver_identifier = 'urn:epc:id:sgln:039999.111111.0'
        eoc.end_point = self._create_endpoint()
        eoc.authentication_info = self._create_auth()
        eoc.save()
        return eoc

    def _create_endpoint(self):
        ep = models.EndPoint()
        ep.urn = getattr(settings, 'TEST_SERVER', 'http://testhost')
        ep.name = 'Test EndPoint'
        ep.save()
        return ep

    def _create_auth(self):
        auth = models.AuthenticationInfo()
        auth.description = 'Unit test auth.'
        auth.username = 'UnitTestUser'
        auth.password = 'UnitTestPassword'
        auth.save()
        return auth

    def _create_bad_criterion(self):
        eoc = EPCISOutputCriteria()
        eoc.name = "Test Criteria"
        eoc.action = "DELETE"
        eoc.event_type = EventType.Transaction.value
        endpoint = self._create_endpoint()
        auth = self._create_auth()
        eoc.end_point = endpoint
        eoc.authentication_info = auth
        eoc.save()
        return eoc

    def _parse_data(self, output_criteria):
        curpath = os.path.dirname(__file__)
        parser = SimpleOutputParser(
            os.path.join(curpath, 'data/epcis.xml'),
            output_criteria
        )
        parser.parse()
        parser.clear_cache()

    def _create_delay_rule(self):
        rule = Rule()
        rule.name = 'delay-rule'
        rule.description = 'a simple delay rule'
        rule.save()
        self.create_delay_step(rule)
        return rule

    def create_delay_step(self, rule, order=1):
        step = Step()
        step.step_class = 'quartet_output.steps.DelayStep'
        step.order = order
        step.name = 'wait 3 seconds'
        step.rule = rule
        step.save()
        param = StepParameter()
        param.step = step
        param.name = 'Timeout Interval'
        param.value = '3'
        param.save()

    def _create_rule(self):
        rule = Rule()
        rule.name = 'output-test'
        rule.description = 'output test rule'
        rule.save()
        return rule

    def _create_filtered_output_step(self, rule, order=1):
        step = Step()
        step.step_class = 'quartet_output.steps.EPCPyYesFilteredEventOutputStep'
        step.order = order
        step.name = 'filtered output step'
        step.rule = rule
        step.save()

    def _create_transport_rule(self):
        rule = Rule()
        rule.name = 'Transport Rule'
        rule.description = 'Attempts to send data using transport step(s).'
        rule.save()
        return rule

    def _create_transport_step(self, rule, put_data='False'):
        step = Step()
        step.rule = rule
        step.order = 1
        step.name = 'Transport'
        step.step_class = 'quartet_output.steps.TransportStep'
        step.description = 'Sends test data.'
        step.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'run-immediately'
        step_parameter.value = 'True'
        step_parameter.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'put-data'
        step_parameter.value = put_data
        step_parameter.save()

    def _create_step(self, rule):
        step = Step()
        step.rule = rule
        step.order = 1
        step.name = 'Output Determination'
        step.step_class = 'quartet_output.steps.OutputParsingStep'
        step.description = 'unit test step'
        step.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'EPCIS Output Criteria'
        step_parameter.value = 'Test Criteria'
        step_parameter.save()
        return step

    def _create_output_steps(self, rule):
        step = Step()
        step.rule = rule
        step.order = 2
        step.name = 'UnpackHierarchies'
        step.step_class = 'quartet_output.steps.UnpackHierarchyStep'
        step.description = 'unit test unpacking step'
        step.save()

    def _create_comm_step(self, rule):
        step = Step()
        step.rule = rule
        step.order = 3
        step.name = 'CreateCommissioning'
        step.step_class = 'quartet_output.steps.AddCommissioningDataStep'
        step.description = 'unit test commissioning step'
        step.save()

    def _create_epcpyyes_step(self, rule, json=False):
        step = Step()
        step.rule = rule
        step.order = 4
        step.name = 'Create EPCIS'
        step.step_class = 'quartet_output.steps.EPCPyYesOutputStep'
        step.description = 'Creates EPCIS XML or JSON and inserts into rule' \
                           'context.'
        step.save()
        if json:
            param = StepParameter.objects.create(
                step=step,
                name='JSON',
                value=True
            )

    def _create_forward_data_step(self, rule):
        step = Step()
        step.rule = rule
        step.order = 4
        step.name = 'Forward Data'
        step.step_class = 'quartet_output.steps.ForwardDataOutputStep'
        step.description = 'Just forwards the data.'
        step.save()

    def _create_task(self, rule):
        task = Task()
        task.rule = rule
        task.name = 'unit test task'
        task.save()
        return task

    def _add_forward_data_step_parameter(self, step: Step):
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'Forward Data'
        step_parameter.value = 'True'
        step_parameter.description = 'Whether or not to construct new data ' \
                                     'or to just forward the data in the ' \
                                     'rule.'
        step_parameter.save()

    def _create_task_step(self, rule, order=5):
        step = Step()
        step.rule = rule
        step.order = order
        step.name = 'Create Output Task'
        step.step_class = 'quartet_output.steps.CreateOutputTaskStep'
        step.description = 'Looks for any EPCIS data on the context and ' \
                           'then, if found, creates a new output task using ' \
                           'the configured Output Rule step parameter.'
        step.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'Output Rule'
        step_parameter.value = 'Transport Rule'
        step_parameter.description = 'The name of the rule to create a new ' \
                                     'task with.'
        step_parameter.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'run-immediately'
        step_parameter.value = 'True'
        step_parameter.description = 'The name of the rule to create a new ' \
                                     'task with.'
        step_parameter.save()
        return step

    def _parse_test_data(self, test_file='data/epcis.xml',
                         parser_type=BusinessEPCISParser,
                         recursive_decommission=False):
        curpath = os.path.dirname(__file__)
        if isinstance(parser_type, BusinessEPCISParser):
            parser = parser_type(
                os.path.join(curpath, test_file),
                recursive_decommission=recursive_decommission
            )
        else:
            parser = parser_type(
                os.path.join(curpath, test_file),
            )
        message_id = parser.parse()
        print(parser.event_cache)
        parser.clear_cache()
        return message_id, parser

    def tearDown(self):
        pass


class TestSFTPTransport(TestQuartetOutput):

    def _create_endpoint(self):
        ep = models.EndPoint()
        ep.urn = 'sftp://testsftphost:22/upload'
        ep.name = 'Test EndPoint SFTP'
        ep.save()
        return ep

    def _create_auth(self):
        auth = models.AuthenticationInfo()
        auth.description = 'Unit test auth.'
        auth.username = 'foo'
        auth.password = 'pass'
        auth.save()
        return auth

    def test_uploaded_file_output(self):
        self._create_good_header_criterion()
        db_rule = self._create_rule()
        db_step = self._create_step(db_rule)
        db_task_step = self._create_task_step(db_rule)
        self._add_forward_data_step_parameter(db_task_step)
        db_step.order = 1
        db_task_step.order = 2
        db_step.save()
        db_task_step.save()
        db_rule2 = self._create_transport_rule()
        self._create_transport_step(db_rule2)
        curpath = os.path.dirname(__file__)
        db_task = self._create_task(db_rule)
        data_path = os.path.join(curpath, 'data/epcis.xml')
        with open(data_path, 'r') as data_file:
            data_out = data_file.read()
            context = execute_rule(data_out.encode(), db_task)
            for event in context.context[
                ContextKeys.FILTERED_EVENTS_KEY.value
            ]:
                self.assertIsInstance(event, sbdheader)
            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY.value]
            execute_queued_task(task_name=task_name)
            task = Task.objects.get(name=task_name)
            self.assertEqual(task.status, 'FINISHED')
            try:
                sftp_client = paramiko.SSHClient()
                sftp_client.set_missing_host_key_policy(
                    paramiko.AutoAddPolicy())
                sftp_client.connect('testsftphost', '22', username='foo',
                                    password='pass', timeout=60)
                sftp = sftp_client.open_sftp()
                remote_file = sftp.open('/upload/' + task_name + '.xml', 'r')
                data_back = remote_file.read()
                self.assertEqual(data_back.decode(), data_out)
            finally:
                sftp_client.close()

class TestMail(TestCase):
    def test_mailto(self):
        urn = 'mailto:hither@this.local?body=send%20current-issue&subject=awesome email'
        mail_info = urlparse(urn)
        mixin = MailMixin()
        print(settings.EMAIL_BACKEND)
        message = mixin.convert_mailto_url(urn)
        message.attach('test.txt', '<test>asdf</test>', 'text/plain')
        message.send(fail_silently=False)
        print(mail_info)
